# encoding: utf-8
import re
import warnings

from mo_dots import listwrap
from mo_future import text, Iterable
from mo_imports import delay_import

from mo_parsing import whitespaces
from mo_parsing.enhancement import (
    Combine,
    Forward,
    Group,
    Suppress,
    ZeroOrMore,
)
from mo_parsing.expressions import MatchFirst, Or
from mo_parsing.results import ParseResults, NO_PARSER
from mo_parsing.tokens import (
    CaselessKeyword,
    CaselessLiteral,
    Keyword,
    NoMatch,
    Literal,
    Empty,
)
from mo_parsing.utils import regex_range, wrap_parse_action

Regex = delay_import("mo_parsing.regex.Regex")


def delimited_list(expr, separator=",", combine=False):
    """
    PARSE DELIMITED LIST OF expr
    Example::

        delimited_list(Word(alphas)).parse_string("aa,bb,cc") # -> ['aa', 'bb', 'cc']
        delimited_list(Word(hexnums), delim=':', combine=True).parse_string("AA:BB:CC:DD:EE") # -> ['AA:BB:CC:DD:EE']
    """
    if combine:
        return Combine(expr + ZeroOrMore(separator + expr, whitespaces.CURRENT))
    else:
        return expr + ZeroOrMore(Suppress(separator) + expr, whitespaces.CURRENT)


def one_of(strs, caseless=False, as_keyword=False):
    """
    Helper to quickly define a set of alternative Literals, and makes
    sure to do longest-first testing when there is a conflict,
    regardless of the input order, but returns
    a `MatchFirst` for best performance.

    Parameters:

     - strs - a string of space-delimited literals, or a collection of
       string literals
     - caseless - (default= ``False``) - treat all literals as caseless
     - as_keyword - (default=``False``) - enforce Keyword-style matching on the
       generated expressions
    """
    if isinstance(caseless, text):
        warnings.warn(
            "More than one string argument passed to one_of, pass "
            "choices as a list or space-delimited string",
            stacklevel=2,
        )

    if caseless:
        isequal = lambda a, b: a.upper() == b.upper()
        masks = lambda a, b: b.upper().startswith(a.upper())
        parseElementClass = CaselessKeyword if as_keyword else CaselessLiteral
    else:
        isequal = lambda a, b: a == b
        masks = lambda a, b: b.startswith(a)
        parseElementClass = Keyword if as_keyword else Literal

    symbols = []
    if isinstance(strs, text):
        symbols = strs.split()
    elif isinstance(strs, Iterable):
        symbols = list(strs)
    else:
        warnings.warn(
            "Invalid argument to one_of, expected string or iterable",
            SyntaxWarning,
            stacklevel=2,
        )
    if not symbols:
        return NoMatch()

    if not as_keyword:
        # if not producing keywords, need to reorder to take care to avoid masking
        # longer choices with shorter ones
        i = 0
        while i < len(symbols) - 1:
            cur = symbols[i]
            for j, other in enumerate(symbols[i + 1 :]):
                if isequal(other, cur):
                    del symbols[i + j + 1]
                    break
                elif masks(cur, other):
                    del symbols[i + j + 1]
                    symbols.insert(i, other)
                    break
            else:
                i += 1

    if caseless or as_keyword:
        return MatchFirst(parseElementClass(sym) for sym in symbols).streamline()

    # CONVERT INTO REGEX
    singles = [s for s in symbols if len(s) == 1]
    rest = list(sorted([s for s in symbols if len(s) != 1], key=lambda s: -len(s)))

    acc = []
    acc.extend(re.escape(sym) for sym in rest)
    if singles:
        acc.append(regex_range("".join(singles)))
    regex = "|".join(acc)

    return Regex(regex).streamline()


LEFT_ASSOC = object()
RIGHT_ASSOC = object()
_no_op = Empty().suppress()


def infix_notation(
    base_expr, spec, lpar=Suppress(Literal("(")), rpar=Suppress(Literal(")"))
):
    """
    :param base_expr: expression representing the most basic element for the
       nested
    :param spec: list of tuples, one for each operator precedence level
       in the expression grammar; each tuple is of the form ``(op_expr,
       numTerms, rightLeftAssoc, parse_action)``, where:

       - op_expr is the mo_parsing expression for the operator; may also
         be a string, which will be converted to a Literal; if numTerms
         is 3, op_expr is a tuple of two expressions, for the two
         operators separating the 3 terms
       - numTerms is the number of terms for this operator (must be 1,
         2, or 3)
       - rightLeftAssoc is the indicator whether the operator is right
         or left associative, using the mo_parsing-defined constants
         ``RIGHT_ASSOC`` and ``LEFT_ASSOC``.
       - parse_action is the parse action to be associated with
         expressions matching this operator expression
    :param lpar: expression for matching left-parentheses
       (default= ``Suppress('(')``)
    :param rpar: expression for matching right-parentheses
       (default= ``Suppress(')')``)
    :return: ParserElement
    """

    all_op = {}

    def norm(op):
        if op == None:
            op = _no_op
        output = all_op.get(id(op))
        if output:
            return output

        def record_self(tok):
            ParseResults(tok.type, tok.start, tok.end, [tok.type.parser_name], [])

        output = whitespaces.CURRENT.normalize(op)
        is_suppressed = isinstance(output, Suppress)
        if is_suppressed:
            output = output.expr
        output = output / record_self
        all_op[id(op)] = is_suppressed, output
        return is_suppressed, output

    op_list = []
    """
    SCRUBBED LIST OF OPERATORS
    * expr - used exclusively for ParseResult(expr, [...]), not used to match
    * op - used to match 
    * arity - same
    * assoc - same
    * parse_actions - same
    """

    for oper_def in spec:
        op, arity, assoc, rest = oper_def[0], oper_def[1], oper_def[2], oper_def[3:]
        parse_actions = list(map(wrap_parse_action, listwrap(rest[0]))) if rest else []
        if arity == 1:
            is_suppressed, op = norm(op)
            if assoc == RIGHT_ASSOC:
                op_list.append((
                    Group(base_expr + op),
                    op,
                    is_suppressed,
                    arity,
                    assoc,
                    parse_actions,
                ))
            else:
                op_list.append((
                    Group(op + base_expr),
                    op,
                    is_suppressed,
                    arity,
                    assoc,
                    parse_actions,
                ))
        elif arity == 2:
            is_suppressed, op = norm(op)
            op_list.append((
                Group(base_expr + op + base_expr),
                op,
                is_suppressed,
                arity,
                assoc,
                parse_actions,
            ))
        elif arity == 3:
            is_suppressed, op = zip(norm(op[0]), norm(op[1]))
            op_list.append((
                Group(base_expr + op[0] + base_expr + op[1] + base_expr),
                op,
                is_suppressed,
                arity,
                assoc,
                parse_actions,
            ))
    op_list = tuple(op_list)

    def record_op(op):
        def output(tokens):
            return ParseResults(NO_PARSER, tokens.start, tokens.end, [(tokens, op)], [])

        return output

    prefix_ops = MatchFirst([
        op / record_op(op)
        for expr, op, is_suppressed, arity, assoc, pa in op_list
        if arity == 1 and assoc == RIGHT_ASSOC
    ])
    suffix_ops = MatchFirst([
        op / record_op(op)
        for expr, op, is_suppressed, arity, assoc, pa in op_list
        if arity == 1 and assoc == LEFT_ASSOC
    ])
    ops = Or([
        op_part / record_op(op_part)
        for op_part in set(
            op_part
            for expr, op, is_suppressed, arity, assoc, pa in op_list
            if arity > 1
            for op_part in (op if isinstance(op, tuple) else [op])
        )
    ])

    def make_tree(tokens, loc, string):
        flat_tokens = list(tokens)
        num = len(op_list)
        op_index = 0
        while len(flat_tokens) > 1 and op_index < num:
            expr, op, is_suppressed, arity, assoc, parse_actions = op_list[op_index]
            if arity == 1:
                if assoc == RIGHT_ASSOC:
                    # PREFIX OPERATOR -3
                    todo = list(reversed(list(enumerate(flat_tokens[:-1]))))
                    for i, (r, o) in todo:
                        if o == op:
                            tok = flat_tokens[i + 1][0]
                            if is_suppressed:
                                result = ParseResults(
                                    expr, tok.start, tok.end, (tok,), []
                                )
                            else:
                                result = ParseResults(
                                    expr, r.start, tok.end, (r, tok), []
                                )
                            break
                    else:
                        op_index += 1
                        continue
                else:
                    # SUFFIX OPERATOR 3!
                    todo = list(enumerate(flat_tokens[1:]))
                    for i, (r, o) in todo:
                        if o == op:
                            tok = flat_tokens[i][0]
                            if is_suppressed:
                                result = ParseResults(
                                    expr, tok.start, tok.end, (tok,), []
                                )
                            else:
                                result = ParseResults(
                                    expr, tok.start, r.end, (tok, r,), []
                                )
                            break
                    else:
                        op_index += 1
                        continue
            elif arity == 2:
                todo = list(enumerate(flat_tokens[1:-1]))
                if assoc == RIGHT_ASSOC:
                    todo = list(reversed(todo))

                for i, (r, o) in todo:
                    if o == op:
                        if is_suppressed:
                            result = ParseResults(
                                expr,
                                flat_tokens[i][0].start,
                                flat_tokens[i + 2][0].end,
                                (flat_tokens[i][0], flat_tokens[i + 2][0]),
                                [],
                            )
                        else:
                            result = ParseResults(
                                expr,
                                flat_tokens[i][0].start,
                                flat_tokens[i + 2][0].end,
                                (flat_tokens[i][0], r, flat_tokens[i + 2][0]),
                                [],
                            )
                        break
                else:
                    op_index += 1
                    continue

            else:  # arity==3
                todo = list(enumerate(flat_tokens[1:-3]))
                if assoc == RIGHT_ASSOC:
                    todo = list(reversed(todo))

                for i, (r0, o0) in todo:
                    if o0 == op[0]:
                        r1, o1 = flat_tokens[i + 3]
                        if o1 == op[1]:
                            seq = [
                                flat_tokens[i][0],
                                flat_tokens[i + 2][0],
                                flat_tokens[i + 4][0],
                            ]
                            s0, s1 = is_suppressed
                            if not s1:
                                seq.insert(2, r1)
                            if not s0:
                                seq.insert(1, r0)

                            result = ParseResults(
                                expr, seq[0].start, seq[-1].end, seq, []
                            )
                            break
                else:
                    op_index += 1
                    continue

            for p in parse_actions:
                result = p(result, -1, string)
            offset = (0, 2, 3, 5)[arity]
            flat_tokens[i : i + offset] = [(result, (expr,))]
            op_index = 0

        result = flat_tokens[0][0]
        result.end = tokens.end
        result.failures = tokens.failures
        return result

    flat = Forward()
    iso = lpar.suppress() + flat + rpar.suppress()
    atom = (base_expr | iso) / record_op(base_expr)
    modified = ZeroOrMore(prefix_ops) + atom + ZeroOrMore(suffix_ops)
    flat << ((modified + ZeroOrMore(ops + modified)) / make_tree).streamline()

    return flat.streamline()



# encoding: utf-8
import re
from collections import OrderedDict

from mo_dots import Null, is_null
from mo_future import text, is_text
from mo_imports import export, expect

from mo_parsing import whitespaces
from mo_parsing.core import ParserElement
from mo_parsing.exceptions import (
    ParseException,
    RecursiveGrammarException,
)
from mo_parsing.results import ParseResults, Annotation
from mo_parsing.utils import (
    Log,
    listwrap,
    empty_tuple,
    regex_iso,
    append_config,
    regex_compile,
    wrap_parse_action,
)
from mo_parsing.utils import MAX_INT, is_forward

Word: object
(
    Token,
    NoMatch,
    Literal,
    Keyword,
    Word,
    CharsNotIn,
    StringEnd,
    Empty,
    Char,
) = expect(
    "Token",
    "NoMatch",
    "Literal",
    "Keyword",
    "Word",
    "CharsNotIn",
    "StringEnd",
    "Empty",
    "Char",
)

_get = object.__getattribute__


class ParseEnhancement(ParserElement):
    """
    Abstract subclass of `ParserElement`, for combining and
    post-processing parsed tokens.
    """

    __slots__ = ["expr"]

    def __init__(self, expr):
        ParserElement.__init__(self)
        self.expr = expr = whitespaces.CURRENT.normalize(expr)
        if is_forward(expr):
            expr.track(self)

    def copy(self):
        output = ParserElement.copy(self)
        output.expr = self.expr
        return output

    def expecting(self):
        if self.expr:
            return OrderedDict(((k, [self]) for k, e in self.expr.expecting().items()))
        else:
            return {}

    def _min_length(self):
        return self.expr.min_length()

    @property
    def whitespace(self):
        return self.expr.whitespace

    def parse_impl(self, string, start, do_actions=True):
        result = self.expr._parse(string, start, do_actions)
        return ParseResults(self, result.start, result.end, [result], result.failures)

    def streamline(self):
        if self.streamlined:
            return self

        expr = self.expr.streamline()
        if expr is self.expr:
            self.streamlined = True
            return self

        if not expr or isinstance(expr, Empty) and not self.is_annotated():
            return Empty()

        output = self.copy()
        output.expr = expr
        output.streamlined = True
        return output

    def check_recursion(self, seen=empty_tuple):
        if self in seen:
            raise RecursiveGrammarException(seen + (self,))
        if self.expr != None:
            self.expr.check_recursion(seen + (self,))

    def __str__(self):
        if self.parser_name:
            return self.parser_name
        return f"{self.__class__.__name__}:({self.expr})"


class LookAhead(ParseEnhancement):
    """
    Lookahead matching of the given parse expression.
    `LookAhead` does *not* advance the parsing position within
    the input string, it only verifies that the specified parse
    expression matches at the current position.  `LookAhead`
    always returns a null token list. If any results names are defined
    in the lookahead expression, those *will* be returned for access by
    name.
    """

    zero_length = True
    __slots__ = []

    def __init__(self, expr):
        ParseEnhancement.__init__(self, expr)

    def parse_impl(self, string, start, do_actions=True):
        # by using self._expr.parse and deleting the contents of the returned ParseResults list
        # we keep any named results that were defined in the FollowedBy expression
        result = self.expr._parse(string, start, do_actions=do_actions)
        result.__class__ = Annotation

        return ParseResults(self, start, start, [result], result.failures)

    def __regex__(self):
        return "*", f"(?={self.expr.__regex__()[1]})"


FollowedBy = LookAhead


class NotAny(LookAhead):
    """
    Lookahead to disallow matching with the given parse expression.
    May be constructed using the '~' operator.
    """

    zero_length = True
    __slots__ = ["regex"]

    def __init__(self, expr):
        super(NotAny, self).__init__(expr)
        prec, pattern = self.expr.__regex__()
        try:
            self.regex = regex_compile(f"(?!{pattern})")
        except Exception as c:
            self.regex = None

    def copy(self):
        output = ParseEnhancement.copy(self)
        output.regex = self.regex
        return output

    def parse_impl(self, string, start, do_actions=True):
        regex = self.regex
        if regex:
            found = regex.match(string, start)
            if found:
                return ParseResults(self, start, start, [], [])
            raise ParseException(self, start, string)
        else:
            try:
                self.expr.parse(string, start, do_actions=True)
                raise ParseException(self, start, string)
            except:
                return ParseResults(self, start, start, [], [])

    def streamline(self):
        output = ParseEnhancement.streamline(self)
        if isinstance(output, Empty):
            return NoMatch()
        if isinstance(output.expr, NoMatch):
            return Empty()
        if isinstance(output.expr, Empty):
            return NoMatch()
        return output

    def expecting(self):
        return {}

    def min_length(self):
        return 0

    def __regex__(self):
        return "*", self.regex.pattern

    def __str__(self):
        if self.parser_name:
            return self.parser_name
        return "~{" + str(self.expr) + "}"


class Many(ParseEnhancement):
    __slots__ = []
    Config = append_config(
        ParseEnhancement, "whitespace", "min_match", "max_match", "end"
    )

    def __init__(
            self,
            expr,
            whitespace=None,
            stop_on=None,
            min_match=0,
            max_match=MAX_INT,
            exact=None,
    ):
        """
        MATCH expr SOME NUMBER OF TIMES (OR UNTIL stop_on IS REACHED
        :param expr: THE EXPRESSION TO MATCH
        :param stop_on: THE PATTERN TO INDICATE STOP MATCHING (NOT REQUIRED IN PATTERN, JUST A QUICK STOP)
        :param min_match: MINIMUM MATCHES REQUIRED FOR SUCCESS (-1 IS INVALID)
        :param max_match: MAXIMUM MATCH REQUIRED FOR SUCCESS (-1 IS INVALID)
        """
        ParseEnhancement.__init__(self, expr)
        if exact is not None:
            min_match = exact
            max_match = exact

        self.set_config(
            whitespace=whitespace or whitespaces.CURRENT,
            min_match=min_match,
            max_match=max_match,
        )
        self.stop_on(stop_on)

    def stop_on(self, ender):
        if ender:
            end = self.parser_config.whitespace.normalize(ender)
            self.set_config(end=regex_compile(end.__regex__()[1]))
        return self

    def _min_length(self):
        if self.parser_config.min_match == 0:
            return 0
        return self.expr.min_length()

    @property
    def whitespace(self):
        return self.parser_config.whitespace

    def parse_impl(self, string, start, do_actions=True):
        acc = []
        end = index = start
        max = self.parser_config.max_match
        stopper = self.parser_config.end
        count = 0
        failures = []
        try:
            while end < len(string) and count < max:
                if end > index:
                    if isinstance(self.expr, LookBehind):
                        index = end
                    else:
                        index = self.parser_config.whitespace.skip(string, end)
                if stopper:
                    if stopper.match(string, index):
                        if self.parser_config.min_match <= count:
                            break
                        else:
                            raise ParseException(
                                self, end, string, msg="found stopper too soon"
                            )
                result = self.expr._parse(string, index, do_actions)
                end = result.end
                if result:
                    acc.append(result)
                    count += 1
        except ParseException as cause:
            if self.parser_config.min_match <= count <= max:
                failures.append(cause)
            else:
                raise ParseException(
                    self,
                    start,
                    string,
                    msg="Not correct amount of matches",
                    cause=cause,
                ) from None
        if count:
            if (
                    count < self.parser_config.min_match
                    or self.parser_config.max_match < count
            ):
                raise ParseException(
                    self,
                    acc[0].start,
                    string,
                    msg=(
                        f"Expecting between {self.parser_config.min_match} and"
                        f" {self.parser_config.max_match} of {self.expr}"
                    ),
                )
            else:
                return ParseResults(self, acc[0].start, acc[-1].end, acc, failures)
        else:
            if not self.parser_config.min_match:
                return ParseResults(self, start, start, [], failures)
            else:
                raise ParseException(
                    self,
                    start,
                    string,
                    msg=f"Expecting at least {self.parser_config.min_match} of {self}",
                )

    def streamline(self):
        if self.streamlined:
            return self
        try:
            expr = self.expr.streamline()
        except Exception as e:
            print(e)
        if (
                self.parser_config.min_match == self.parser_config.max_match
                and not self.is_annotated()
        ):
            if self.parser_config.min_match == 0:
                return Empty()
            elif self.parser_config.min_match == 1:
                return expr

        if self.expr is expr:
            self.streamlined = True
            return self
        if expr.is_annotated() or not isinstance(expr, Empty):
            output = self.copy()
            output.expr = expr
            output.streamlined = True
            return output
        return Empty()

    def __regex__(self):
        end = self.parser_config.end.pattern if self.parser_config.end else None
        prec, regex = self.expr.__regex__()
        regex = regex_iso(prec, regex, "*")

        if self.parser_config.max_match == MAX_INT:
            if self.parser_config.min_match == 0:
                suffix = "*"
            elif self.parser_config.min_match == 1:
                suffix = "+"
            else:
                suffix = "{" + text(self.parser_config.min_match) + ",}"
        elif self.parser_config.min_match == self.parser_config.max_match:
            if self.parser_config.min_match == 1:
                suffix = ""
            else:
                suffix = "{" + text(self.parser_config.min_match) + "}"
        else:
            suffix = (
                    "{"
                    + text(self.parser_config.min_match)
                    + ","
                    + text(self.parser_config.max_match)
                    + "}"
            )

        if end:
            return "+", regex + suffix + end
        else:
            return "*", regex + suffix

    def __call__(self, name):
        if not name:
            return self

        for e in [self.expr]:
            if isinstance(e, ParserElement) and e.token_name == name:
                Log.error(
                    "can not set token name, already set in one of the other"
                    " expressions"
                )

        return ParseEnhancement.__call__(self, name)

    def __str__(self):
        if self.parser_name:
            return self.parser_name
        return f"{self.__class__.__name__}:({self.expr})"


class OneOrMore(Many):
    """Repetition of one or more of the given expression.

    Parameters:
     - expr - expression that must match one or more times
     - stop_on - (default= ``None``) - expression for a terminating sentinel
          (only required if the sentinel would ordinarily match the repetition
          expression)
    """

    __slots__ = []

    def __init__(self, expr, whitespace=None, stop_on=None):
        Many.__init__(self, expr, whitespace, stop_on, min_match=1, max_match=MAX_INT)

    def __str__(self):
        if self.parser_name:
            return self.parser_name
        return "{" + text(self.expr) + "}+"


class ZeroOrMore(Many):
    """Optional repetition of zero or more of the given expression.

    Parameters:
     - expr - expression that must match zero or more times
     - stop_on - (default= ``None``) - expression for a terminating sentinel
          (only required if the sentinel would ordinarily match the repetition
          expression)

    Example: similar to `OneOrMore`
    """

    __slots__ = []

    def __init__(self, expr, whitespace=None, stop_on=None):
        Many.__init__(self, expr, whitespace, stop_on=stop_on, min_match=0, max_match=MAX_INT)

    def parse_impl(self, string, start, do_actions=True):
        try:
            return super(ZeroOrMore, self).parse_impl(string, start, do_actions)
        except ParseException as pe:
            return ParseResults(self, start, start, [], [pe])

    def __str__(self):
        if self.parser_name:
            return self.parser_name

        return "(" + text(self.expr) + ")*"


class Optional(Many):
    """Optional matching of the given expression.

    Parameters:
     - expr - expression that must match zero or more times
     - default (optional) - value to be returned if the optional expression is not found.
    """

    __slots__ = []
    Config = append_config(Many, "default_value")

    def __init__(self, expr, whitespace=None, default=None):
        Many.__init__(self, expr, whitespace, stop_on=None, min_match=0, max_match=1)
        self.set_config(default_value=listwrap(default))

    def parse_impl(self, string, start, do_actions=True):
        try:
            results = self.expr._parse(string, start, do_actions)
            return ParseResults(self, results.start, results.end, [results], [])
        except ParseException as pe:
            return ParseResults(
                self, start, start, self.parser_config.default_value, [pe]
            )

    def __str__(self):
        if self.parser_name:
            return self.parser_name

        return "[" + text(self.expr) + "]"


class SkipTo(ParseEnhancement):
    """Token for skipping over all undefined text until the matched expression is found."""

    __slots__ = []
    Config = append_config(ParseEnhancement, "include", "fail", "ignore", "whitespace")

    def __init__(self, expr, include=False, ignore=None, fail_on=None, engine_=None):
        """
        :param expr: target expression marking the end of the data to be skipped
        :param include: if True, the target expression is also parsed
          (the skipped text and target expression are returned as a 2-element list).
        :param ignore: used to define grammars (typically quoted strings and
          comments) that might contain false matches to the target expression
        :param fail_on: define expressions that are not allowed to be
          included in the skipped test; if found before the target expression is found,
          the SkipTo is not a match
        """
        ParseEnhancement.__init__(self, expr)

        self.set_config(
            include=include,
            fail=whitespaces.CURRENT.normalize(fail_on),
            ignore=ignore,
            whitespace=engine_ or whitespaces.CURRENT,
        )
        self.parser_name = str(self)

    def min_length(self):
        return 0

    def parse_impl(self, string, start, do_actions=True):
        instrlen = len(string)
        fail = self.parser_config.fail
        ignore = self.parser_config.ignore

        loc = start
        while loc <= instrlen:
            skip_end = loc
            loc = before_end = self.parser_config.whitespace.skip(string, loc)
            if fail:
                # break if fail_on expression matches
                try:
                    fail._parse(string, loc)
                    break
                except:
                    pass

            if ignore:
                # advance past ignore expressions
                while 1:
                    try:
                        loc = ignore._parse(string, loc).end
                        skip_end = loc
                        loc = before_end = self.parser_config.whitespace.skip(
                            string, loc
                        )
                    except ParseException:
                        break
            try:
                loc = self.expr._parse(string, loc, do_actions=False).end
            except ParseException:
                # no match, advance loc in string
                loc += 1
            else:
                # matched skipto expr, done
                break
        else:
            # ran off the end of the input string without matching skipto expr, fail
            raise ParseException(self, start, string)

        # build up return values
        end = loc
        skiptext = string[start:skip_end]
        skip_result = []
        if skiptext:
            skip_result.append(skiptext)

        if self.parser_config.include:
            end_result = self.expr._parse(string, before_end, do_actions)
            skip_result.append(end_result)
            return ParseResults(self, start, end, skip_result, [])
        else:
            return ParseResults(self, start, before_end, skip_result, [])


class Forward(ParserElement):
    """Forward declaration of an expression to be defined later -
    used for recursive grammars, such as algebraic infix notation.
    When the expression is known, it is assigned to the ``Forward``
    variable using the '<<' operator.

    Note: take care when assigning to ``Forward`` not to overlook
    precedence of operators.

    Specifically, '|' has a lower precedence than '<<', so that::

        fwd_expr << a | b | c

    will actually be evaluated as::

        (fwd_expr << a) | b | c

    thereby leaving b and c out as parseable alternatives.  It is recommended that you
    explicitly group the values inserted into the ``Forward``::

        fwd_expr << (a | b | c)

    Converting to use the '<<=' operator instead will avoid this problem.

    See `ParseResults.pprint` for an example of a recursive
    parser created using ``Forward``.
    """

    __slots__ = ["expr", "used_by", "_str", "_reg", "_eng"]

    def __init__(self, expr=Null):
        ParserElement.__init__(self)
        self.expr = None
        self.used_by = []

        self._str = None  # avoid recursion
        self._reg = None  # avoid recursion
        self._eng = False
        if expr:
            self << whitespaces.CURRENT.normalize(expr)

    def copy(self):
        output = ParserElement.copy(self)
        output.expr = self
        output._str = None
        output._reg = None
        output._eng = False

        output.used_by = []
        return output

    @property
    def name(self):
        return self.type.expr.token_name

    def track(self, expr):
        self.used_by.append(expr)

    def __lshift__(self, other):
        self._str = ""
        if is_forward(self.expr):
            return self.expr << other

        while is_forward(other):
            other = other.expr
        norm = whitespaces.CURRENT.normalize(other)
        if norm is None:
            Log.error("problem")
        self.expr = norm.streamline()
        self.check_recursion()
        return self

    def add_parse_action(self, action):
        self.parse_action.append(wrap_parse_action(action))
        return self

    def leave_whitespace(self):
        with whitespaces.NO_WHITESPACE:
            output = self.copy()
            output.expr = self.expr.leave_whitespace()
            return output

    def streamline(self):
        if not self.expr or self.expr.streamlined:
            return self

        self.expr = self.expr.streamline()
        self.check_recursion()
        return self

    def check_recursion(self, seen=empty_tuple):
        if self in seen:
            raise RecursiveGrammarException(seen + (self,))
        if self.expr != None:
            self.expr.check_recursion(seen + (self,))

    def min_length(self):
        if self.min_length_cache is None and self.expr:
            self.min_length_cache = 0  # BREAK CYCLE
            try:
                return self.expr.min_length()
            finally:
                self.min_length_cache = None
        return 0

    @property
    def whitespace(self):
        try:
            if self._eng:
                return None
        except Exception as cause:
            Log.error("", cause=cause)

        # Avoid infinite recursion by setting a temporary
        self._eng = True
        try:
            return self.expr.whitespace
        finally:
            self._eng = False

    def parse_impl(self, string, loc, do_actions=True):
        try:
            result = self.expr._parse(string, loc, do_actions)
            return ParseResults(
                self, result.start, result.end, [result], result.failures
            )
        except Exception as cause:
            if is_null(self.expr):
                Log.warning(
                    "Ensure you have assigned a ParserElement (<<) to this Forward",
                    cause=cause,
                )
            raise cause from None

    def __regex__(self):
        if self._reg or not self.expr:
            return None

        try:
            self._reg = True
            return self.expr.__regex__()
        finally:
            self._reg = None

    def __str__(self):
        if self.parser_name:
            return self.parser_name

        if self._str:
            return self._str

        # Avoid infinite recursion
        self._str = "Forward: ..."
        try:
            self._str = "Forward: " + text(self.expr)[:1000]
        except Exception:
            pass
        return self._str

    def __call__(self, name):
        output = self.copy()
        output.token_name = name
        return output


class TokenConverter(ParseEnhancement):
    """
    Abstract subclass of `ParseExpression`, for converting parsed results.
    """

    __slots__ = []

    def __regex__(self):
        return self.expr.__regex__()


class Combine(TokenConverter):
    """
    Converter to concatenate all matching tokens to a single string.
    """

    Config = append_config(TokenConverter, "separator")

    def __init__(self, expr, separator=""):
        super(Combine, self).__init__(expr.streamline())
        self.set_config(separator=separator)

    def parse_impl(self, string, start, do_actions=True):
        result = self.expr.parse_impl(string, start, do_actions=do_actions)
        output = ParseResults(
            self,
            start,
            result.end,
            [result.as_string(sep=self.parser_config.separator)],
            result.failures,
        )
        return output

    def streamline(self):
        if self.streamlined:
            return self

        expr = self.expr.streamline()
        if expr is self.expr:
            self.streamlined = True
            return self
        return Combine(expr, self.parser_config.separator)

    def expecting(self):
        return OrderedDict((k, [self]) for k in self.expr.expecting().keys())

    def min_length(self):
        return self.expr.min_length()

    def __regex__(self):
        return self.expr.__regex__()

    def __str__(self):
        return text(self.expr)


class Group(TokenConverter):
    """
    MARK A CLOSED PARSE RESULT
    """

    __slots__ = []

    def __init__(self, expr):
        ParserElement.__init__(self)
        self.expr = whitespaces.CURRENT.normalize(expr)

    def is_annotated(self):
        return True


class Dict(Group):
    """
    Convert a list of tuples [(name, v1, v2, ...), ...]
    int dict-like lookup     {name: [v1, v2, ...], ...}

    mo-parsing uses the names of the ParserElement to name ParseResults,
    but this is a static naming scheme. Dict allows dynamic naming;
    Effectively defining new named ParserElements (called Annotations)
    at parse time
    """

    __slots__ = []

    def __init__(self, expr):
        Group.__init__(self, expr)
        self.parse_action.append(_dict_post_parse)


class OpenDict(TokenConverter):
    """
    Same as Dict, but not grouped: Open to previous (or subsequent) name: value pairs
    """

    __slots__ = []

    def __init__(self, expr):
        TokenConverter.__init__(self, expr)
        self.parse_action.append(_dict_post_parse)


def _dict_post_parse(tokens, loc, string):
    acc = tokens.tokens
    for a in list(acc):
        for tok in list(a):
            if not tok:
                continue
            if is_text(tok):
                new_tok = Annotation(tok, a.start, a.end, [])
            else:
                kv = list(tok)
                key = kv[0]
                value = kv[1:]
                new_tok = Annotation(text(key), tok.start, tok.end, value)
            acc.append(new_tok)

    return tokens


class Suppress(TokenConverter):
    """
    Converter for ignoring the results of a parsed expression.
    """

    __slots__ = []

    def __init__(self, expr):
        TokenConverter.__init__(self, expr)
        self.parse_action.append(_suppress_post_parse)

    def suppress(self):
        return self

    def __regex__(self):
        return self.expr.__regex__()

    def __str__(self):
        if self.parser_name:
            return self.parser_name
        return text(self.expr)


def _suppress_post_parse(tokens, start, string):
    return ParseResults(tokens.type, tokens.start, tokens.end, [], tokens.failures)


class PrecededBy(LookAhead):
    """
    Lookbehind matching of the given parse expression.
    ``PrecededBy`` does not advance the parsing position within the
    input string, it only verifies that the specified parse expression
    matches prior to the current position.  ``PrecededBy`` always
    returns a null token list, but if a results name is defined on the
    given expression, it is returned.

    Parameters:

     - expr - expression that must match prior to the current parse
       location
     - retreat - (default= ``None``) - (int) maximum number of characters
       to lookbehind prior to the current parse location

    If the lookbehind expression is a string, Literal, Keyword, or
    a Word or CharsNotIn with a specified exact or maximum length, then
    the retreat parameter is not required. Otherwise, retreat must be
    specified to give a maximum number of characters to look back from
    the current parse position for a lookbehind match.
    """

    zero_length = True
    __slots__ = []
    Config = append_config(ParseEnhancement, "retreat", "exact")

    def __init__(self, expr, retreat=None):
        super(PrecededBy, self).__init__(expr)
        expr = self.expr

        if isinstance(expr, (Literal, Keyword, Char)):
            self.set_config(retreat=expr.min_length(), exact=True)
        elif isinstance(expr, (Word, CharsNotIn)):
            self.set_config(retreat=expr.min_length(), exact=False)
        elif expr.__class__.zero_length:
            self.set_config(retreat=0, exact=True)
        else:
            self.set_config(retreat=expr.min_length(), exact=False)

    def parse_impl(self, string, start=0, do_actions=True):
        if self.parser_config.exact:
            loc = start - self.parser_config.retreat
            if loc < 0:
                raise ParseException(self, start, string)
            ret = self.expr._parse(string, loc)
        else:
            # retreat specified a maximum lookbehind window, iterate
            test_expr = self.expr + StringEnd()
            instring_slice = string[:start]
            last_cause = ParseException(self, start, string)

            for offset in range(self.parser_config.retreat, start + 1):
                try:
                    ret = test_expr._parse(instring_slice, start - offset)
                    break
                except ParseException as cause:
                    last_cause = cause
            else:
                raise last_cause
        # return empty list of tokens, but preserve any defined results names

        ret.__class__ = Annotation
        return ParseResults(self, start, start, [ret], [])

    def __regex__(self):
        if self.parser_config.exact:
            return "*", f"(?<={self.expr.__regex__()[1]})"
        raise NotImplemented()


LookBehind = PrecededBy


export("mo_parsing.core", SkipTo)
export("mo_parsing.core", Many)
export("mo_parsing.core", ZeroOrMore)
export("mo_parsing.core", OneOrMore)
export("mo_parsing.core", Optional)
export("mo_parsing.core", NotAny)
export("mo_parsing.core", Suppress)
export("mo_parsing.core", Group)

export("mo_parsing.results", Group)
export("mo_parsing.results", Dict)
export("mo_parsing.results", Suppress)

export("mo_parsing.whitespaces", Empty)

export("mo_parsing.utils", Many)



# encoding: utf-8
import sys

from mo_future import text

from mo_parsing.core import ParserElement
from mo_parsing.utils import (
    quote,
    lineno,
    col,
    stack_depth,
    quote as plain_quote, ParseException,
)

DEBUGGING = False


class Debugger(object):
    def __init__(self, silent=False):
        self.previous_parse = None
        self.was_debugging = False
        self.parse_count = 0
        self.max_stack_depth = 0
        self.silent = silent

    def __enter__(self):
        global DEBUGGING
        self.was_debugging = DEBUGGING
        DEBUGGING = True
        self.previous_parse = ParserElement._parse
        ParserElement._parse = _debug_parse(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        global DEBUGGING
        ParserElement._parse = self.previous_parse
        DEBUGGING = self.was_debugging


def _debug_parse(debugger):
    def debug_parse(self, string, start, do_actions=True):
        if not debugger.silent:
            _try(self, start, string)
        loc = start
        try:
            debugger.parse_count += 1
            debugger.max_stack_depth = stackdepth()
            tokens = self.parse_impl(string, loc, do_actions)
        except ParseException as cause:
            self.parser_config.fail_action and self.parser_config.fail_action(
                self, start, string, cause
            )
            if not debugger.silent:
                fail(self, start, string, cause)
            raise ParseException(self, start, string, cause=cause) from None

        if self.parse_action and (do_actions or self.parser_config.callDuringTry):
            try:
                for fn in self.parse_action:
                    tokens = fn(tokens, start, string)
            except Exception as cause:
                fail(self, start, string, cause)
                raise
        if not debugger.silent:
            match(self, loc, tokens.end, string, tokens)

        return tokens

    return debug_parse


def _try(expr, start, string):
    print(
        "  Attempt "
        + quote(string, start)
        + " at loc "
        + text(start)
        + "(%d,%d)" % (lineno(start, string), col(start, string))
        + " for "
        + " " * stack_depth()
        + text(expr)[:300]
    )


def match(expr, start, end, string, tokens):
    print(
        "> Matched "
        + quote(string[start:end])
        + "between "
        + f"[{start}, {end}] for"
        + " " * stack_depth()
        + text(expr)
        + " -> "
        + str(tokens)
    )


def fail(expr, start, string, cause):
    quoted = plain_quote(text(cause))
    print("  Except  " + quoted)


def quote(value, start=0, length=12):
    return (plain_quote(value[start : start + length - 2]) + (" " * length))[:length]


def stackdepth():
    """
    return depth of stack at call point
    """
    frame = sys._getframe(1)
    depth = 0
    while frame:
        depth += 1
        frame = frame.f_back
    return depth
